use std::fs::File;
use std::io::{Error, ErrorKind, BufReader, BufWriter, Read, Write};

use alchemy::*;



///
const LOADER_MAJOR_VERSION: u8 = 0;

///
const LOADER_MINOR_VERSION: u8 = 1;

///
const HEADER_BYTE_SIZE: usize = 6;

///
const MAGIC_NUMBER: &'static str = "MSNL";



///
pub struct Entry
{
   ///
   name: String,

   ///
   type_hint: String,

   ///
   path: String
}

///
pub struct Loader
{
   ///
   entries: Vec<Entry>
}



impl Entry
{
   ///
   pub fn new(entry_name: String, entry_type: String, entry_path: String)
      -> Entry
   {
      Entry
      {
         name: entry_name,
         type_hint: entry_type,
         path: entry_path
      }
   }

   ///
   pub fn get_name(&self) -> &str
   {
      self.name.as_str()
   }

   ///
   pub fn get_type(&self) -> &str
   {
      self.type_hint.as_str()
   }

   ///
   pub fn get_path(&self) -> &str
   {
      self.path.as_str()
   }
}

impl Transmutable for Entry
{
   fn to_bytes(&self, endianess: Endianess) -> Vec<u8>
   {
      let mut buffer: Vec<u8>;

      // Create the new buffer to hold the data from this type.
      buffer = Vec::new();

      // Add the buffer data of the three paths to the data buffer.
      buffer.append(&mut self.name.to_bytes(endianess));
      buffer.append(&mut self.type_hint.to_bytes(endianess));
      buffer.append(&mut self.path.to_bytes(endianess));

      // Return the data buffer.
      buffer.shrink_to_fit();
      buffer
   }

   fn from_bytes(buffer: &[u8], endianess: Endianess) -> Entry
   {
      let buffer_size: usize;
      let name: String;
      let type_hint: String;
      let path: String;
      let mut skip_size: usize;

      buffer_size = buffer.len() - 1;

      skip_size = 0;
      name = String::from_bytes(&buffer[skip_size..buffer_size], endianess);

      skip_size = skip_size + get_byte_size_of_string(&name);
      type_hint = String::from_bytes(&buffer[skip_size..buffer_size],
                                     endianess);

      skip_size = skip_size + get_byte_size_of_string(&type_hint);
      path = String::from_bytes(&buffer[skip_size..buffer_size], endianess);

      Entry::new(name, type_hint, path)
   }

   fn determine_byte_size(&self) -> usize
   {
      get_byte_size_of_string(&self.name) +
      get_byte_size_of_string(&self.type_hint) +
      get_byte_size_of_string(&self.path)
   }
}

impl Loader
{
   ///
   pub fn new() -> Loader
   {
      Loader
      {
         entries: Vec::new()
      }
   }

   ///
   pub fn from_file(file: File) -> Loader
   {
      let mut loader: Loader;

      // Create a new loader.
      loader = Loader
      {
         entries: Vec::new()
      };

      // Read the given file.
      match loader.read_from_file(file)
      {
         Ok(_) =>
         {
         }

         Err(error) =>
         {
            error!("{}", error);
         }
      }

      // Return the loader.
      loader
   }

   ///
   pub fn add_resource_entry(&mut self, name: String,
                             type_hint: String, path: String)
   {
      let entry: Entry;

      entry = Entry::new(name, type_hint, path);
      self.entries.push(entry);
   }

   ///
   pub fn get_resource_entries(&self) -> &Vec<Entry>
   {
      &self.entries
   }

   /// Write the resource entries out to a file.
   pub fn write_to_file(&self, file: File) -> ::std::io::Result<()>
   {
      let mut data: Vec<u8>;
      let mut writer: BufWriter<File>;

      // Create the writer for the file.
      writer = BufWriter::new(file);

      // Create a Vector for the data of this loader data.
      data = Vec::new();

      // Write the header of the file.
      data.extend_from_slice(MAGIC_NUMBER.as_bytes());
      data.append(&mut LOADER_MAJOR_VERSION.to_bytes(Endianess::Little));
      data.append(&mut LOADER_MINOR_VERSION.to_bytes(Endianess::Little));

      // Write all the entries of the file.
      for entry in self.entries.iter()
      {
         data.append(&mut entry.to_bytes(Endianess::Little));
      }

      // Try to write all the data to the file.
      try!(writer.write_all(data.as_slice()));

      Ok(())
   }

   /// Read resource entries from a file.
   fn read_from_file(&mut self, file: File) -> ::std::io::Result<()>
   {
      let mut pos: usize;
      let mut data: Vec<u8>;
      let mut reader: BufReader<File>;
      let data_size: usize;
      let data_slice: &[u8];
      let mut entry: Entry;
      debug!("Parsing loader manifest.");

      // Create a new reader of the file.
      reader = BufReader::new(file);

      // Try to read in all of the file.
      data = Vec::new();
      try!(reader.read_to_end(&mut data));
      data_slice = data.as_slice();
      data_size = data_slice.len() - 1;

      // Check the header information.
      pos = 0;
      try!(self.parse_header(&data_slice[pos..HEADER_BYTE_SIZE]));
      pos = pos + HEADER_BYTE_SIZE;

      // Read in each entry.
      while pos <= data_size
      {
         debug!("\tParsing entry.");
         entry = Entry::from_bytes(&data_slice[pos..data_size],
                                   Endianess::Little);
         pos = pos + entry.determine_byte_size();

         self.entries.push(entry);
      }

      debug!("Loader manifest parsed successfully.");
      Ok(())
   }

   fn parse_header(&self, header: &[u8]) -> ::std::io::Result<()>
   {
      let magic_number: &[u8];

      // Make sure the header starts with the magic number.
      magic_number = MAGIC_NUMBER.as_bytes();
      for (i, byte) in magic_number.iter().enumerate()
      {
         if *byte != header[i]
         {
            return Err(Error::new(ErrorKind::InvalidData,
                                  "Header magic number mismatch."));
         }
      }

      // Make sure the version is correct.
      if header[4] != LOADER_MAJOR_VERSION || header[5] != LOADER_MINOR_VERSION
      {
         return Err(Error::new(ErrorKind::InvalidData,
                               "Header wrong version."));
      }

      // This is a good header.
      debug!("Header parsed successfully.");
      Ok(())
   }
}
